@@ -4,10 +4,10 @@
 
 ;; Author: Michał Krzywkowski <k.michal@zoho.com>
 ;; Keywords: languages, tools
-;; Package-Version: 0.1.0
-;; Package-Revision: v0.1.0-0-g2f31aae3548a
+;; Package-Version: 0.2.0
+;; Package-Revision: v0.2.0-0-g0f50d49dff71
 ;; Homepage: https://github.com/mkcms/flymake-x
-;; Package-Requires: ((emacs "27.1") (flymake "1.0"))
+;; Package-Requires: ((emacs "28.1") (flymake "1.0"))
 
 ;; This program is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -43,6 +43,14 @@
 ;;
 ;; Call `flymake-x-setup' once, and modify `flymake-x-checkers'.  See it's
 ;; documentation for details.
+;;
+;; You can also load the included \\='flymake-x-sample-checkers\\=' library
+;; which contains a few predefined checkers.
+;;
+;; For easily defining new checkers, a helpful command is provided to display
+;; the checker output, found diagnostics and other information:
+;; `flymake-x-debug'.  It should be called with the name of a checker or it's
+;; definition, as described in `flymake-x-checkers'.
 
 ;;; Code:
 
@@ -91,6 +99,11 @@ functions.")
 (cl-defgeneric flymake-x-stop (checker)
   "Stop CHECKER.")
 
+(cl-defmethod slot-unbound ((_object flymake-x-checker) _class _slot-name _fn)
+  "Override `slot-unbound' error for all checker objects.
+We override it to return nil in that case."
+  nil)
+
 
 ;; Process checker
 
@@ -131,7 +144,7 @@ functions.")
               (flymake-x-checker-name checker))))
 
 (cl-defmethod initialize-instance
-  :after ((instance flymake-x-process-checker) _slots) ;lint:ignore
+  :after ((instance flymake-x-process-checker) _slots)
   "Constructor for flymake-x process checker class."
   (when (functionp (flymake-x-checker-command instance))
     (setf (flymake-x-checker-command instance)
@@ -182,6 +195,40 @@ functions.")
     (when (process-live-p process)
       (kill-process process))))
 
+(eval-and-compile
+  (defvar flymake-x--rx-base-group 10))
+
+(defmacro flymake-x--rx-let-eval (&rest body)
+  "Evaluate BODY with extended rx forms supported by flymake-x.
+Special variables are available in BODY:
+
+- \\='line-group\\='
+- \\='column-group\\='
+- \\='message-group\\='
+- \\='file-group\\='
+
+These variables hold the regexp group numbers that would be matched by
+flymake-x patterns."
+  (declare (indent 0) (debug (body)))
+  (let* ((line flymake-x--rx-base-group)
+         (column (1+ line))
+         (message (1+ column))
+         (file (1+ message)))
+    `(let ((line-group ,line)
+           (column-group ,column)
+           (message-group ,message)
+           (file-group ,file))
+       ;; Prevent warning about unused variables.
+       (ignore line-group)
+       (ignore column-group)
+       (ignore message-group)
+       (ignore file-group)
+       (rx-let-eval '((line (group-n ,line (one-or-more digit)))
+                      (column (group-n ,column (one-or-more digit)))
+                      (message (&rest pat) (group-n ,message pat))
+                      (file (group-n ,file (zero-or-more nonl))))
+         ,@body))))
+
 (defun flymake-x--patterns-search (patterns)
   "Search for rx PATTERNS in current buffer and return a match.
 
@@ -207,24 +254,23 @@ LINE is either nil or the line number matched by the pattern,
 COLUMN is either nil or the column matched by the pattern,
 MESSAGE is either nil or the message matched by the pattern."
   (save-match-data
-    (rx-let-eval '((line (group-n 10 (one-or-more digit)))
-                   (column (group-n 11 (one-or-more digit)))
-                   (message (&rest pat) (group-n 12 pat))
-                   (file (group-n 13 (zero-or-more nonl))))
+    (flymake-x--rx-let-eval
       (let (retval start)
         (while (and (not (eobp)) (null retval))
           (setq start (point))
           (setq retval
-                (cl-loop for pattern in patterns
-                         for i from 0
-                         if (looking-at (rx-to-string `(seq ,@pattern)))
-                         return (list i
-                                      (when-let* ((line (match-string 10)))
-                                        (string-to-number line))
-                                      (when-let* ((column (match-string 11)))
-                                        (string-to-number column))
-                                      (when-let* ((message (match-string 12)))
-                                        message))))
+                (cl-loop
+                 for pattern in patterns
+                 for i from 0
+                 if (looking-at (rx-to-string `(seq ,@pattern)))
+                 return
+                 (list i
+                       (when-let* ((line (match-string line-group)))
+                         (string-to-number line))
+                       (when-let* ((column (match-string column-group)))
+                         (string-to-number column))
+                       (when-let* ((message (match-string message-group)))
+                         message))))
           (if retval
               (goto-char (match-end 0))
             (forward-line 1)
@@ -285,7 +331,7 @@ MESSAGE is either nil or the message matched by the pattern."
   :documentation "A process checker that accepts the input file via stdin.")
 
 (cl-defmethod flymake-x-start
-  :after ((checker flymake-x-pipe-checker)) ;lint:ignore
+  :after ((checker flymake-x-pipe-checker))
   "Start a pipe CHECKER."
   (when-let* ((process (flymake-x-checker-process checker)))
     (process-send-region process (point-min) (point-max))
@@ -303,7 +349,7 @@ to that file is appended to the checker's command line.  After
 the checker process exits, the file is deleted.")
 
 (cl-defmethod flymake-x-start
-  :before ((checker flymake-x-temp-file-checker)) ;lint:ignore
+  :before ((checker flymake-x-temp-file-checker))
   "Start a temp-file CHECKER."
   (when-let* ((buffer-file (buffer-file-name)))
     (let* ((directory (file-name-directory buffer-file))
@@ -317,7 +363,7 @@ the checker process exits, the file is deleted.")
             (concat (flymake-x-checker-command checker) " " tempfile)))))
 
 (cl-defmethod flymake-x-stop
-  :after ((checker flymake-x-temp-file-checker)) ;lint:ignore
+  :after ((checker flymake-x-temp-file-checker))
   "Stop a temp-file CHECKER."
   (when-let* ((file (flymake-x-checker-temp-file checker)))
     (when (file-exists-p file) (delete-file file))))
@@ -490,23 +536,165 @@ If THIS-BUFFER-ONLY, enable it only in this buffer instead."
   (add-hook 'flymake-diagnostic-functions #'flymake-x--start-checkers
             nil this-buffer-only))
 
-;; This code is only evaluated when linting.  It is used to ignore certain
-;; unfixable warnings on older Emacs versions.
-;;
-;;lint: (progn
-;;lint:   (require 'checkdoc)
-;;lint:   (defvar lint-orig-checkdoc-create-error-function
-;;lint:     checkdoc-create-error-function)
-;;lint:   (defun lint-checkdoc-create-error-function (text start end &rest args)
-;;lint:     "Create checkdoc error for TEXT START END ARGS."
-;;lint:     (save-excursion
-;;lint:       (goto-char start)
-;;lint:       (goto-char (line-beginning-position))
-;;lint:       (when (looking-at ".*;.*\\blint:ignore\\b")
-;;lint:         (setq text (concat "__ignore: " text)))
-;;lint:       (apply lint-orig-checkdoc-create-error-function text start end args)))
-;;lint:   (setq checkdoc-create-error-function
-;;lint:         #'lint-checkdoc-create-error-function))
+
+;; Debugging
+
+(defun flymake-x-debug (checker)
+  "Debug CHECKER: run it and show a buffer displaying helpful information.
+CHECKER should be either a symbol or a checker definition as described
+in `flymake-x-checkers'.  If it is a symbol, then the checker definition
+is looked up in `flymake-x-checkers' using that symbol.
+
+The checker is run for the current buffer.
+The displayed information includes:
+- Process exit status
+- Command that was run
+- Found diagnostics and details
+- Compiled error patterns."
+  (interactive
+   (list
+    (intern
+     (completing-read "Debug checker in this buffer: "
+                      (mapcar #'symbol-name (mapcar #'car flymake-x-checkers))
+                      nil t))))
+  (when (symbolp checker)
+    (setq checker (assoc checker flymake-x-checkers)))
+
+  (flymake-x--cleanup)
+
+  (let* ((flymake-x-checkers (list checker))
+         (buf (current-buffer))
+         (goto-pt (lambda (point)
+                    (lambda (&rest _args)
+                      (with-current-buffer buf
+                        (pop-to-buffer (current-buffer))
+                        (goto-char point)))))
+         process
+         stderr-buffer
+         diagnostics
+         sentinel
+         process-output)
+    (flymake-x--start-checkers (lambda (diags) (setq diagnostics diags)))
+    (unless flymake-x-buffer-checkers
+      (error (concat
+              "No checkers were created for the current buffer;"
+              " is the checker defined for another major mode?")))
+    (setq checker (car flymake-x-buffer-checkers))
+
+    (with-current-buffer (get-buffer-create "*flymake-x debug*")
+      (setq process (flymake-x-checker-process checker))
+      (when process
+        (setq sentinel (process-sentinel process))
+        (set-process-sentinel
+         process
+         (lambda (&rest args)
+           (with-current-buffer (flymake-x-process-output-buffer checker)
+             (setq process-output (buffer-string)))
+           (apply sentinel args))))
+
+      (ignore-errors (kill-buffer "*flymake-x debug process output*"))
+
+      (let ((inhibit-read-only t))
+        (setq buffer-read-only t)
+        (setq buffer-undo-list t)
+        (setq truncate-lines t)
+        (erase-buffer)
+
+        (setq stderr-buffer (format " *stderr of flymake-x-%s*"
+                                    (flymake-x-checker-name checker)))
+
+        (when process
+          (save-current-buffer
+            (while (accept-process-output process))
+            (sit-for 0.1)))
+
+        (let ((standard-output (current-buffer)))
+          (insert-button "[Show flymake log buffer]"
+                               'action (lambda (&rest _args)
+                                         (flymake-switch-to-log-buffer)))
+                (insert "\n")
+          (insert (format "Checker: %s\n" (flymake-x-checker-name checker)))
+          (insert (format "Type: %s\n" (eieio-object-class checker)))
+          (insert (format "Diagnostics from stderr: %s\n"
+                          (if (flymake-x-checker-use-stderr-p checker)
+                              "yes" "no")))
+          (insert (format "Lines start from 0: %s\n"
+                          (if (flymake-x-checker-lines-start-from-0-p checker)
+                              "yes" "no")))
+          (insert
+           (format "Columns start from 1: %s\n"
+                   (if (flymake-x-checker-columns-start-from-1-p checker)
+                       "yes" "no")))
+          (insert (format "Command: %S\n" (flymake-x-checker-command checker)))
+          (if (not process)
+              (insert "Error: Process was not created\n")
+            (insert (format "Process exit status: %s\n"
+                            (process-exit-status process)))
+            (insert (format "Output: %s chars " (length process-output)))
+            (insert-button
+             "[Show process output]" 'action
+             (lambda (&rest _args)
+               (with-current-buffer
+                   (get-buffer-create "*flymake-x debug process output*")
+                 (let ((inhibit-read-only t))
+                   (setq buffer-read-only t)
+                   (setq buffer-undo-list t)
+                   (erase-buffer)
+                   (insert process-output)
+                   (pop-to-buffer (current-buffer))
+                   (special-mode)))))
+            (insert "\n")
+            (insert "Stderr: ")
+            (insert-button
+             "[Show stderr buffer]" 'action
+             (lambda (&rest _args) (pop-to-buffer stderr-buffer)))
+            (insert "\n"))
+          (insert "Error patterns:\n")
+          (cl-loop for pat in (flymake-x-checker-error-patterns checker)
+                   do (progn (insert "\n")
+                             (insert "    ")
+                             (prin1 pat)
+                             (insert "\n")
+                             (insert "    Which is:\n")
+                             (insert
+                              (format "    %S"
+                                      (flymake-x--rx-let-eval
+                                        (rx-to-string `(seq ,@(cadr pat))))))
+                             (insert "\n")))
+          (insert "\n")
+
+          (insert (format "Diagnostics found: %s\n" (length diagnostics)))
+          (pcase-dolist
+              (`(,i . ,diag) (cl-loop for x in diagnostics
+                                      for i from 0
+                                      collect (cons i x)))
+            (insert "\n")
+            (insert (format "    Index: %d\n" i))
+            (insert (format "    Type: %s\n"
+                            (flymake-diagnostic-type diag)))
+
+            (insert (format "    Beg: %s " (flymake-diagnostic-beg diag)))
+            (insert-button "[Goto]" 'action
+                           (funcall goto-pt (flymake-diagnostic-beg diag)))
+            (insert "\n")
+            (insert (format "    End: %s " (flymake-diagnostic-end diag)))
+            (insert-button "[Goto]" 'action
+                           (funcall goto-pt (flymake-diagnostic-end diag)))
+            (insert "\n")
+
+            (insert (format "    Text: %s\n"
+                            (format "%S" (flymake-diagnostic-text diag)))))))
+      (special-mode)
+      (button-mode)
+      (font-lock-add-keywords nil
+                              '(("\\(:error\\)" 1 'error prepend)
+                                ("^\\(Error:\\)" 1 'error prepend)
+                                ("\\(:warning\\)" 1 'warning prepend)
+                                ("\\(:note\\)" 1 'flymake-note-echo prepend)))
+      (setq-local font-lock-keywords-only nil)
+      (font-lock-fontify-region (point-min) (point-max))
+      (goto-char (point-min))
+      (pop-to-buffer (current-buffer)))))
 
 (provide 'flymake-x)
 ;;; flymake-x.el ends here
